/**
 * DimpCore.js - Dimp UI application logic.
 *
 * $Horde: dimp/js/src/DimpCore.js,v 1.285 2007/07/30 04:31:37 slusarz Exp $
 *
 * Copyright 2005-2007 The Horde Project (http://www.horde.org/)
 *
 * See the enclosed file COPYING for license information (GPL). If you
 * did not receive this file, see http://www.fsf.org/copyleft/gpl.html.
 */

var baseWindow = false, DMenu,
    /* Trick some Horde js into thinking this is the parent Horde window. */
    frames = { horde_main: true };

/* DimpCore object. */
var DimpCore = {
    // Vars used and defaulting to null/false:
    //   viewport, folder, uid, ppID, lastrow, folderswitch,
    //   inAjaxCallback, pollPE, currentButton, isvisible, offset
    compose_id: 1,
    drags: $H(),
    remove_gc: [],
    selected: [],
    server_error: 0,
    showPreview: DimpVar.preview_pref,

    buttons: [ 'button_reply', 'button_forward', 'button_spam', 'button_ham', 'button_delete' ],

    sfiltersfolder: { sf_all: 'all', sf_current: 'current' },
    sfilters: { sf_msgall: 'all', sf_from: 'from', sf_to: 'to', sf_subject: 'subject' },

    debug: function(label, e) {
        if (DimpVar.debug) {
            if (typeof e.message == 'string') {
                e = e.message;
            }
            alert(label + ': ' + e);
        }
    },

    // Message selection functions
    select: function(id, opts) {
        opts = opts || $H();
        var count = this.selectedCount(), uflag = [];
        if (count && !opts.add) {
            this.clearSelected();
        }

        id.each(function(s) {
            s = parseInt(s);
            if (opts.add) {
                if (!this.isSelected(s)) {
                    this.selected.push(s);
                    ++count;
                }
            } else {
                this.selected = [ s ];
                count = 1;
            }

            if (this.viewport.inBuffer(s)) {
                uflag.push(s);
                if (this.lastrow != -1) {
                    this.lastrow = parseInt(this.viewport.getRowById(s).rownum);
                }
            }
        }.bind(this));

        if (uflag.size()) {
            this.viewport.updateFlag(uflag, 'selectedRow', true, false);
        }
        this.toggleButtons();

        if (!opts.nopreview && $('previewPane').visible()) {
            if (opts.right) {
                this.clearPreviewPane();
            } else {
                this.initPreviewPane();
            }
        }
    },

    deselect: function(id, opts) {
        opts = opts || $H();
        this.selected = this.selected.without(id);
        if (id.size()) {
            this.viewport.updateFlag(id, 'selectedRow', false, false);
        }

        var count = this.selectedCount();
        if (!count) {
            this.lastrow = null;
        }

        if (!opts.clearall) {
            this.toggleButtons();
            if (opts.right || !count) {
                this.clearPreviewPane();
            } else if ((count == 1) && $('previewPane').visible()) {
                var row = this.viewport.getRowById(this.selected.first());
                this.loadPreviewPane({ mbox: row.mbox, uid: row.uid });
            }
        }
    },

    ctrlSelect: function(id, e) {
        id = parseInt(id);
        var args = $H({ right: this.isRightClick(e) });
        if (this.isSelected(id)) {
            this.deselect([ id ], args);
        } else {
            this.select([ id ], args.merge($H({ add: true })));
        }
    },

    shiftSelect: function(id, e) {
        id = parseInt(id);
        var args = { add: true, right: e && this.isRightClick(e) },
            lastrow = this.lastrow;
        this.select([ id ], args);
        var count = this.selectedCount();
        if (count == 1) {
            return;
        }

        DMenu.close();

        // Logic: link the current id with the last clicked id.
        var row = this.viewport.getRowById(id);
        var bounds = [ row.rownum, lastrow ];
        var start = bounds.min();

        this.processShiftSelect(start, bounds.max() - start + 1);
    },

    processShiftSelect: function(s, l) {
        var ids = this.viewport.getRows(s, l);
        if (ids.size() == l) {
            this.shiftSelectAdd(ids.pluck('uid'));
        } else {
            Effect.Appear('folderLoading', { duration: 0.2 });
            this.doAction('GetMsgUIDSlice', { start: s, length: l, folder: this.folder }, [], this.shiftSelectCallback.bind(this));
        }
    },

    selectAll: function() {
        DMenu.close();
        this.processShiftSelect(1, this.viewport.getMetaData('total_rows'));
    },

    shiftSelectCallback: function(r) {
        if (r.response.uids) {
            this.shiftSelectAdd(r.response.uids.parseRangeString());
        }
        Effect.Fade('folderLoading', { duration: 0.2 });
    },

    shiftSelectAdd: function(ids) {
        var tmp = this.lastrow;
        this.lastrow = -1;
        this.select(ids, { add: true });
        this.lastrow = tmp;
    },

    selectedCount: function() {
        return this.selected.length;
    },

    clearSelected: function() {
        this.deselect(this.selected, { clearall: true });
    },

    isSelected: function(id) {
        return (this.selected.indexOf(id) != -1);
    },

    getSelectedRows: function() {
        return this.selected.collect(function(s) {
            return { uid: s, mbox: this.folder };
        }.bind(this));
    },

    // absolute = Is num an absolute row number - from 1 -> page_size (true) -
    //            or a relative change from the current selected value (false)
    //            If no current selected value, or more than 1 current
    //            selected value, the first message in the current viewport is
    //            selected.
    moveSelected: function(num, absolute) {
        var curr_row = (this.selectedCount() == 1) ? this.viewport.getRowById(this.selected.first()) : null;

        if (absolute) {
            if (!this.viewport.getMetaData('total_rows')) {
                return;
            }
            var row = this.viewport.getRows(num, 1);
            if (!row.length) {
                this.offset = num;
                this.viewport.updateContent(num - 1);
                return;
            }
        } else {
            if (num == 0) {
                return;
            }

            if (curr_row === null) {
                row = this.viewport.getRows(this.viewport.currentOffset() + 1, 1);
            } else {
                var curr = curr_row.rownum + num;
                curr = (num > 1) ? Math.min(curr, this.viewport.getMetaData('total_rows')) : Math.max(curr, 1);
                var row = this.viewport.getRows(curr, 1);
            }
        }

        if (row.length) {
            row = row.first();
            if (!curr_row || row.uid != curr_row.uid) {
                this.viewport.positionById(row.uid);
                this.select([ row.uid ], { nopreview: true });
                // Pause for 300ms before loading preview message.
                setTimeout(function() { if (this.selectedCount() == 1 && this.selected.first() == row.uid) { this.go('msg:' + this.folder + ':' + row.uid, row); } }.bind(this), 300);
            }
        }
    },
    // End message selection functions

    go: function(location, data) {
        data = data || null;
        this.show(location, data);
        dhtmlHistory.add(location, data);
        return false;
    },

    show: function(location, data) {
        if (location == 'imp') {
            this.go('folder:INBOX');
            return;
        }

        if (location.indexOf('compose:') != -1) {
            return;
        }

        if (location.indexOf('msg:') != -1) {
            if (data && this.viewport) {
                this.select([ data.uid ]);
                this.loadPreviewPane(data);
                return;
            }

            // If data is null then we came directly to this page or the user
            // hit Reload. Load the folder view and then go to the specified
            // message.
            var separator = location.indexOf(':', 4);
            this.uid = location.substring(separator + 1);
            location = 'folder:' + location.substring(4, separator);
            // Now fall through to the isFolder check below.
        }

        if (location.indexOf('folder:') != -1) {
            $('dimpmain_portal').hide();
            $('dimpmain_folder').show();
            this.folder = location.substring(7);
            this.highlightSidebar(this.getFolderId(this.folder));
            this.loadFolder();
            return;
        }

        $('dimpmain_folder').hide();
        $('dimpmain_portal').show();

        if (location.indexOf('app:') != -1) {
            var app = location.substr(4);
            this.highlightSidebar('app' + app);
            this.iframeContent(location, (data) ? data : DimpVar.app_urls[app]);
            return;
        }

        switch (location) {
        case 'portal':
            $('dimpmain_portal').update(DimpText.loading);
            this.highlightSidebar('appportal');
            this.setTitle(DimpText.portal);
            this.doAction('ShowPortal', $H(), [], this.portalCallback.bind(this));
            break;

        case 'options':
            this.highlightSidebar('appoptions');
            this.setTitle(DimpText.prefs);
            this.showOptions();
            break;
        }
    },

    /* 'action' -> if action begins with a '*', the exact string will be used
     *  instead of sending the action to the IMP handler. */
    doAction: function(action, parameters, uids, callback, opts) {
        opts = opts || $H();
        action = (action.length && action.charAt(0) == '*')
            ? action.substring(1)
            : DimpVar.URI_IMP + '/' + action;
        opts.parameters = this.createParamList([ $H(parameters).toQueryString(), ((uids && uids.length) ? ('uid=' + uids.toRangeString()) : '') ]);
        if (callback) {
            opts.onComplete = function(transport, object) {
                this.doActionComplete(transport, callback);
            }.bind(this);
        }
        opts.onException = function(request, exception) { this.debug('onException', exception); }.bind(this);
        opts.onFailure = function(transport, object) { this.debug('onFailure', transport) }.bind(this);
        new Ajax.Request(action, opts);
    },

    doActionComplete: function(request, callback) {
        this.inAjaxCallback = true;
        var error = false, r = $H();

        if (!request.responseText || !request.responseText.length) {
            error = true;
        } else {
            try {
                r = request.responseText.evalJSON(true);
            } catch (e) {
                this.debug('doActionComplete', e);
                error = true;
            }
        }

        if (typeof callback == 'function') {
            try {
                callback(r);
            } catch (e) {
                this.debug('doActionComplete', e);
            }
        }

        if (error) {
            if (++this.server_error == 3) {
                r = { msgs: [ { type: 'horde.error', message: DimpText.ajax_timeout } ], msgs_auto: true };
            }
        } else {
            if (this.server_error >= 3) {
                r.msgs.push({ type: 'horde.success', message: DimpText.ajax_recover });
            }
            this.server_error = 0;
        }

        if (r.msgs && r.msgs_auto) {
            this.showNotifications(r.msgs);
        }

        // Check if the current folder needs refresh.
        if (this.viewport) {
            this.viewport.checkCacheCallback(r);
        }

        this.inAjaxCallback = false;
    },

    createParamList: function(p) {
        p = p || [];
        p.push(DimpVar.SESSION_ID);
        return p.findAll(function(s) { return s.length; }).join('&');
    },

    setTitle: function(title) {
        document.title = DimpVar.name + ' :: ' + title;
    },

    highlightSidebar: function(id) {
        // Folder bar may not be fully loaded yet.
        if ($('foldersLoading').visible()) {
            setTimeout(function() { this.highlightSidebar(id); }.bind(this), 200);
            return;
        }

        $('sidebarPanel').getElementsByClassName('on').each(function(e) {
            $(e).removeClassName('on');
        });

        var elt = $(id);
        if (!elt) {
            return;
        }
        if (elt.tagName.toUpperCase() != 'LI') {
            if (!elt.parentNode) {
                return;
            }
            elt = $(elt.parentNode);
        }
        elt.addClassName('on');

        // Make sure all subfolders are expanded
        elt.ancestors().each(function(n) {
            if (n.hasClassName('subfolders')) {
                this.toggleSubFolder(n.id.substring(3), 'exp');
            }
        }.bind(this));
    },

    showNotifications: function(msgs, doc) {
        if (!msgs.size()) {
            return;
        }

        doc = doc || document;

        var alerts = $('alerts');
        if (!alerts) {
            alerts = $(Builder.node('DIV', { id: 'alerts' }));
            doc.body.appendChild(alerts);
        }

        msgs.each(function(m) {
            switch (m.type) {
            case 'dimp.timeout':
                if (parent.frames.horde_main) {
                    parent.location = DimpVar.timeout_url;
                } else {
                    window.location = DimpVar.timeout_url;
                }
                break;

            case 'horde.error':
            case 'horde.message':
            case 'horde.success':
            case 'horde.warning':
            case 'imp.reply':
            case 'imp.forward':
            case 'imp.redirect':
                var div = $(Builder.node('DIV', { className: m.type.replace('.', '-') }, [ m.message.unescapeHTML() ]));
                var clickdiv = div;
                alerts.appendChild(div);

                // IE6 has a bug that does not allow the body of a div to be
                // clicked to trigger an onclick event for that div (it only
                // seems to be an issue if the div is overlaying an element
                // that itself contains an image).  However, the alert box
                // normally displays over the message list, and we use several
                // graphics in the default message list layout, so we see this
                // buggy behavior 99% of the time.  The workaround is to
                // overlay the div with a like sized div containing a clear
                // gif, which tricks IE into the correct behavior.
                if (DimpVar.is_ie6) {
                    var iefix = $(Builder.node('DIV', { className: 'ie6alertsfix' }));
                    Position.clone(div, iefix, { setLeft: false, setTop: false });
                    clickdiv = iefix;
                    iefix.appendChild(div.remove());
                    alerts.appendChild(iefix);
                }

                clickdiv.observe('click', function() {
                    Effect.Fade(div, { duration: 1.5,
                                       afterFinish: this.removeAlert });
                }.bind(this));

                if (m.type != 'horde.error') {
                    setTimeout(function() {
                        Effect.Fade(div, { duration: 1.5,
                                           afterFinish: this.removeAlert });
                    }.bind(this), 3000);
                }
            }
        }.bind(this));
    },

    removeAlert: function(effect) {
        try {
            var elt = $(effect.element);
            var parent = elt.up();
            if (parent) {
                parent.removeChild(elt);
                if (!parent.immediateDescendants().size() &&
                    parent.hasClassName('ie6alertsfix')) {
                    parent.remove();
                }
            }
        } catch (e) {
            this.debug('removeAlert', e);
        }
    },

    checkQuota: function() {
        this.doAction('chunkContent', { chunk: 'quota' }, [], this.setQuota);
    },

    setQuota: function(r) {
        if (r.response) {
            $('quota').update(r.response.chunk);
        }
    },

    iframeContent: function(name, location) {
        if (name === null) {
            name = location;
        }

        var container = $('dimpmain_portal');
        if (!container) {
            this.showNotifications([ { type: 'horde.error', message: 'Invalid page structure.' } ]);
            return;
        }

        container.update('<iframe class="iframe" src="' + location + '" id="iframe' + name + '" frameborder="0"></iframe>');
    },

    msgWindow: function(args) {
        this.updateUnseenUID(args.uid, args.folder, 0);
        var url = DimpVar.message_url;
        url += ((url.indexOf('?') == -1) ? '?' : '&') +
               $H({ folder: args.folder,
                    thismailbox: args.folder,
                    uid: args.uid }).toQueryString();
        this.popupWindow(url, 'msgview');
        return false;
    },

    compose: function(type, args) {
        args = args || $H();
        var url = DimpVar.compose_url;
        if (type) {
            var params = { type: type };
            $H(args).keys().each(function(prop) {
                if (prop == 'folder') {
                    params.thismailbox = args[prop];
                }
                params[prop] = args[prop];
            });
            url += ((url.indexOf('?') == -1) ? '?' : '&') + $H(params).toQueryString();
        }
        this.popupWindow(url, '');
    },

    popupWindow: function(url, name) {
        if (!window.open(url, name, 'width=' + DimpVar.popup_width + ',height=' + DimpVar.popup_height + ',status=1,scrollbars=yes,resizable=yes')) {
            this.showNotifications([ { type: 'horde.warning', message: DimpText.popup_block } ]);
        }
    },

    closePopup: function() {
        // Mozilla bug/feature: it will not close a browser window
        // automatically if there is code remaining to be performed (or, at
        // least, not here) unless the mouse is moved or a keyboard event
        // is triggered after the callback is complete. (As of FF 2.0.0.3 and
        // 1.5.0.11).  So wait for the callback to complete before attempting
        // to close the window.
        if (this.inAjaxCallback) {
            setTimeout(function() { this.closePopup(); }.bind(this), 25);
        } else {
            window.close();
        }
    },

    composeMailbox: function(type) {
        DMenu.close();
        if (!this.selectedCount()) {
            return;
        }
        this.getSelectedRows().each(function(s) {
            this.compose(type, { folder: s.mbox, uid: s.uid });
        }.bind(this));
    },

    loadFolder: function() {
        this.clearSelected();
        this.toggleButtons();
        this.searchfilterClose();
        this.clearPreviewPane();

        $('folderName').update(DimpText.loading);
        this.folderswitch = true;

        if (!this.viewport) {
            this.createViewPort();
        }

        this.viewport.loadView(this.folder, { folder: this.folder, thismailbox: this.folder }, this.uid);
    },

    createViewPort: function() {
        this.viewport = new ViewPort({
            content_container: 'msgList',
            empty_container: 'msgList_empty',
            error_container: 'msgList_error',
            fetch_action: 'ListMessages',
            cachecheck_action: 'CheckCache',
            msgid_prefix: 'msg',
            template: message_list_template,
            buffer_pages: DimpVar.buffer_pages,
            limit_factor: DimpVar.limit_factor,
            viewport_wait: DimpVar.viewport_wait,
            split_pane: 'previewPane',
            splitbar: 'splitBar',
            content_class: 'msglist',
            row_class: 'msgRow',
            maxHeight: function() {
                return getInnerHeight() - Position.page($('msgList'))[1];
            }.bind(this),
            ajaxRequest: function(action, params, callback, opts) {
                this.doAction(action, params, [], callback, opts);
            }.bind(this),
            onMouseScroll: function() {
                DMenu.close();
            },
            onScrollIdle: function() {
                this.setMessageListTitle();
            }.bind(this),
            onViewChange: function() {
                this.remove_gc = this.remove_gc.concat(this.viewport.visibleRows());
                this.viewport.clear();
            }.bind(this),
            onContent: function(rows) {
                var thread = ((this.viewport.getMetaData('sortby') == DimpVar.sortthread) && this.viewport.getMetaData('thread'));
                rows.each(function(r) {
                    var rid = r.id.substring(3);
                    var row = this.viewport.getRowById(rid);
                    this.viewport.updateFlag([ rid ], 'selectedRow', this.isSelected(rid), false);
                    // Add thread graphics
                    if (thread && thread[row.uid]) {
                        var o, subj = r.down('.msgSubject'), u = thread[row.uid];
                        var tn = subj.firstChild;
                        for (var i = 0, length = u.length; i < length; ++i) {
                            o = $('thread_img_' + u.charAt(i)).cloneNode(false);
                            o.id = '';
                            subj.insertBefore(o, tn);
                        }
                    }
                    // Add attachment graphics
                    if (row.atc) {
                        var size = r.down('.msgSize');
                        var tn = size.firstChild;
                        var n = $('atc_img_' + row.atc).cloneNode(false);
                        n.id = '';
                        size.insertBefore(n, tn);
                    }
                }.bind(this));
            }.bind(this),
            onComplete: function() {
                if (this.uid) {
                    if (this.viewport.positionById(this.uid)) {
                        this.select([ this.uid ]);
                    }
                    this.uid = null;
                } else if (this.offset) {
                    this.select([ this.viewport.getRows(this.offset, 1).first().uid ]);
                    this.offset = null;
                }

                // 'label' will not be set if there has been an error
                // retrieving data from the server.
                var l = this.viewport.getMetaData('label');
                if (l) {
                    $('folderName').update(l);
                }
                if (this.folderswitch) {
                    this.folderswitch = false;

                    if (this.folder == DimpVar.spam_folder) {
                        if (this.buttons.indexOf('button_spam') != -1) {
                            $('button_spam').up().hide();
                            $('ctx_msg_spam').hide();
                        }
                        if (this.buttons.indexOf('button_ham') != -1) {
                            $('button_ham').up().show();
                            $('ctx_msg_ham').show();
                        }
                    } else {
                        if (this.buttons.indexOf('button_spam') != -1) {
                            $('button_spam').up().show();
                            $('ctx_msg_spam').show();
                        }
                        if (this.buttons.indexOf('button_ham') != -1) {
                            $('button_ham').up().hide();
                            $('ctx_msg_ham').hide();
                        }
                    }

                    $('msgFromLink').update(this.viewport.getMetaData('specialfolder') ? DimpText.totext : DimpText.from);
                    this.setSortColumns();
                    if (l) {
                        $('sf_current').update(l);
                    }
                }
                // If sorting by thread, we need to force redrawing all rows
                // if the viewport changes since threadlevel may have changed
                // in the meantime.
                this.viewport.forceUpdate(this.viewport.getMetaData('sortby') == DimpVar.sortthread);
                this.setFolderLabel(this.folder, this.viewport.getMetaData('unseen'));
                this.setMessageListTitle();
                this.updateTitle();
                Effect.Fade('folderLoading', { duration: 0.2 });
            }.bind(this),
            onFetch: function() {
                Effect.Appear('folderLoading', { duration: 0.2 });
            },
            onWait: function() {
                this.showNotifications([ { type: 'horde.warning', message: DimpText.listmsg_wait } ]);
            }.bind(this),
            onFail: function() {
                this.showNotifications([ { type: 'horde.error', message: DimpText.listmsg_timeout } ]);
                Effect.Fade('folderLoading', { duration: 0.2 });
            }.bind(this),
            onFirstContent: function() {
                this.clearPreviewPane();
                $('msgList').observe('dblclick', this.handleMsgListDblclick.bindAsEventListener(this));
                $('msgList').observe('mouseover', this.handleMsgListMouseover.bindAsEventListener(this));
            }.bind(this),
            onRemoveRows: function(r) {
                this.updateUnseenStatus(r.first().mbox, r.findAll(function(row) {
                    return row.bg.match(new RegExp("(^|\\s)unseen(\\s|$)"));
                }).size() * -1);
            }.bind(this),
            onClearRows: function(r) {
                r.each(function(row) {
                    if (row.id) {
                        this.removeMouseEvents(row.id);
                    }
                }.bind(this));
            }.bind(this),
            onBeforeResize: function() {
                this.isvisible = (this.selectedCount() == 1) && (this.viewport.visible(this.selected.first()) == 0);
            }.bind(this),
            onAfterResize: function() {
                if (this.isvisible) {
                    this.viewport.positionById(this.selected.first());
                }
            }.bind(this)
        });

        // DIMP's version of Droppables is smart enough to ignore Draggable
        // items that don't need any droppable listeners.
        $('splitBar').nodrop = true;

        // Initialize preview pane status.
        this.viewport.showSplitPane(this.showPreview);

        // Set up viewport filter events.
        var mf = $('msgList_filter');
        this.viewport.addFilter(mf, 'SearchMessages', this.addSearchfilterParams.bind(this));
        mf.observe('keyup', this.searchfilterStatus.bind(this));
        mf.observe('focus', this.searchfilterOnFocus.bind(this));
        mf.observe('blur', this.searchfilterOnBlur.bind(this));

        // Resize message pane on window size change.
        Event.observe(window, 'resize', this.viewport.onResize.bind(this.viewport));

        // Add loading graphic to message list pane.
        var img = $($('loading_img').down().cloneNode(false)).hide();
        img.id = 'folderLoading';
        $('msgListLoading').appendChild(img);
    },

    handleMsgListDblclick: function(e) {
        DMenu.close();
        var elt = this.getMsgRow(e);
        if (!elt) {
            return;
        }
        var row = this.viewport.getRowById(elt.id.substring(3));
        var hash = { folder: row.mbox, uid: row.uid };
        return (row.draft) ? this.compose('resume', hash) : this.msgWindow(hash);
    },

    handleMsgListMouseover: function(e) {
        var elt = this.getMsgRow(e);
        if (!elt) {
            return;
        }
        this.addMouseEvents(elt.id, this.viewport.getRowById(elt.id.substring(3)).menutype);
    },

    getMsgRow: function(e) {
        var elt = $(Event.element(e));
        if (elt && !elt.hasClassName('msgRow')) {
            elt = elt.up('.msgRow');
        }
        return elt;
    },

    updateTitle: function() {
        var label;
        if ($('qoptions').visible()) {
            label = DimpText.search + ' :: ' + this.viewport.getMetaData('total_rows') + ' ' + DimpText.resfound;
        } else {
            var elt = $(this.getFolderId(this.folder));
            if (elt) {
                var unseen = elt.readAttribute('u');
                label = elt.readAttribute('l');
                if (unseen > 0) {
                    label += ' (' + unseen + ')';
                }
            } else {
                label = this.viewport.getMetaData('label');
            }
        }
        this.setTitle(label);
    },

    isRightClick: function(e) {
        if (e.which) {
            return (e.which == 3);
        } else if (e.button) {
            return (e.button == 2);
        }
        return false;
    },

    messageClick: function(r, e) {
        var args = { right: this.isRightClick(e) };
        this.select([ r ], args);
        if (!args.right) {
            DMenu.close();
            var row = this.viewport.getRowById(r);
            return this.go('msg:' + row.mbox + ':' + row.uid, row);
        }
    },

    sort: function(sortby, e) {
        // Don't change sort if we are past the sortlimit
        if (this.viewport.getMetaData('sortlimit')) {
            return;
        }

        // See if we're switching between two sorts on the same
        // column.
        var clicked = Event.element(e || window.event), s;
        if (clicked.nodeType == 3) {
            clicked = clicked.parentNode;
        }
        if (clicked.tagName.toLowerCase() == 'small') {
            // Swap the major and minor sorts on the column.
            sortby = clicked.readAttribute('sortby');
            var pn = clicked.up(), text = clicked.getText();
            clicked.setAttribute('sortby', pn.readAttribute('sortby'));
            pn.setAttribute('sortby', sortby);
            clicked.setText(pn.getText());
            pn.setText(text);
        }
        if (sortby == this.viewport.getMetaData('sortby')) {
            s = { sortdir: (this.viewport.getMetaData('sortdir') ? 0 : 1) };
            this.viewport.setMetaData({ sortdir: s.sortdir });
        } else {
            s = { sortby: sortby };
            this.viewport.setMetaData({ sortby: s.sortby });
            if (sortby == DimpVar.sortthread) {
                this.viewport.forceUpdate(true);
            }
        }
        this.setSortColumns();
        this.viewport.reload(s);
    },

    setSortColumns: function() {
        var sortby = this.viewport.getMetaData('sortby'), st = $('st');
        var tsort = (sortby == DimpVar.sortthread);

        st.update(tsort ? DimpText.thread : DimpText.subject);
        st.setAttribute('sortby', tsort ? DimpVar.sortthread : DimpVar.sortsubject);
        var s = $(Builder.node('SMALL', { id: 'st_small', sortby: (tsort ? DimpVar.sortsubject : DimpVar.sortthread) }, [ tsort ? DimpText.subject : DimpText.thread ]));
        st.appendChild(s);

        $$('.item').each(function(h) {
            $(h).immediateDescendants().each(function(node) {
                if (!node.tagName || node.tagName.toLowerCase() != 'div') {
                    return;
                }
                var setsort = false;
                node = $(node);
                if (node.hasClassName('msgFrom')) {
                    setsort = (sortby == DimpVar.sortfrom);
                } else if (node.hasClassName('msgSubject')) {
                    setsort = (sortby == DimpVar.sortsubject || sortby == DimpVar.sortthread);
                } else if (node.hasClassName('msgDate')) {
                    setsort = (sortby == DimpVar.sortdate);
                } else if (node.hasClassName('msgSize')) {
                    setsort = (sortby == DimpVar.sortsize);
                }

                if (setsort) {
                    this.isSort(node);
                } else {
                    this.notSort(node);
                }
            }.bind(this));
        }.bind(this));
        if (this.viewport.getMetaData('sortlimit')) {
            $('st_small').hide();
        } else {
            $('st_small').visible();
        }
    },

    toggleButtons: function() {
        var disable = (this.selectedCount() == 0);
        this.buttons.each(function(b) {
            if (disable) {
                $(b).up().addClassName('disabled');
            } else {
                $(b).up().removeClassName('disabled');
            }
            DMenu.disable(b + '_img', true, disable);
        });
    },

    isSort: function(node) {
        var sortdir = this.viewport.getMetaData('sortdir');
        node.removeClassName(sortdir ? 'sortdown' : 'sortup');
        node.addClassName(sortdir ? 'sortup' : 'sortdown');
    },

    notSort: function(node) {
        node.removeClassName('sortup');
        node.removeClassName('sortdown');
    },

    loadPreviewPane: function(data) {
        var pp = $('previewPane');
        if (!pp.visible()) {
            return;
        }
        if (this.ppID &&
            this.ppID.uid == data.uid &&
            this.ppID.mbox == data.mbox) {
            return;
        }
        this.ppID = data;
        this.updateUnseenUID(data.uid, data.mbox, 0);

        // Clear tooltip events.
        this.remove_gc = this.remove_gc.concat($('previewPane').getElementsBySelector('a'));

        var img = $('loading_img').down().cloneNode(false);
        img.id = 'msgLoading';
        $A(pp.childNodes).each(function(n) {
            pp.removeChild(n);
        });
        pp.appendChild(img);

        this.doAction('ShowPreview', { uid: data.uid, folder: data.mbox }, [], this.loadPreviewPaneCallback.bind(this));
    },

    loadPreviewPaneCallback: function(r) {
        if (!r.response || this.selectedCount() != 1) {
            this.clearPreviewPane();
            return;
        }
        r = r.response;
        if (r.index == this.selected.first() && r.folder == this.folder) {
            var pp = $('previewPane');
            pp.update(r.msg);
            this.buildAddressLinks(pp);
            ToolTips.attachBehavior();
        }
    },

    initPreviewPane: function() {
        if (this.selectedCount() != 1) {
            this.clearPreviewPane();
        } else {
            this.loadPreviewPane({ mbox: this.folder, uid: this.selected.first() });
        }
    },

    clearPreviewPane: function() {
        var p = $('previewPane');
        if (p.visible()) {
            // Clear tooltip events.
            this.remove_gc = this.remove_gc.concat($('previewPane').getElementsBySelector('a'));

            var d = $('previewInfoText').cloneNode(true);
            d.id = 'previewInfo';
            d.show();
            p.update('');
            p.appendChild(d);
            this.ppID = null;
        }
    },

    updateUnseenUID: function(uid, mbox, setflag) {
        var r = this.viewport.getRowById(uid);
        if (!r) {
            return false;
        }

        var unseen = 0,
            unseenset = r.bg.match(new RegExp("(^|\\s)unseen(\\s|$)"));
        if ((setflag && unseenset) || (!setflag && !unseenset)) {
            return false;
        }

        if (setflag) {
            this.viewport.updateFlag([ r.uid ], 'unseen', true, true);
            ++unseen;
        } else {
            this.viewport.updateFlag([ r.uid ], 'unseen', false, true);
            --unseen;
        }

        return this.updateUnseenStatus(mbox, unseen);
    },

    updateUnseenStatus: function(mbox, change) {
        if (change == 0) {
            return false;
        }
        var unseen = parseInt($(this.getFolderId(mbox)).readAttribute('u')) + change;
        this.viewport.setMetaData({ unseen: unseen });
        this.setFolderLabel(mbox, unseen);
        this.updateTitle();
        return true;
    },

    setMessageListTitle: function() {
        var rows = this.viewport.getMetaData('total_rows');
        if (rows > 0) {
            var offset = this.viewport.currentOffset();
            $('msgHeader').update(DimpText.messages + ' ' + (offset + 1) + ' - ' + (Math.min(offset + this.viewport.getPageSize(), rows)) + ' ' + DimpText.of + ' ' + rows);
        } else {
            $('msgHeader').update(DimpText.nomessages);
        }
    },

    setFolderLabel: function(f) {
        var fid = this.getFolderId(f), unseen;
        var elt = $(fid);
        if (!elt || !elt.hasAttribute('u')) {
            return;
        }

        if (arguments.length == 2) {
            unseen = Math.max(arguments[1], 0);
            elt.setAttribute('u', unseen);
        } else {
            unseen = parseInt(elt.readAttribute('u'));
        }

        $(fid + '_label').update((unseen > 0) ?
            '<strong>' + elt.readAttribute('l') + '</strong> <span class="count" dir="ltr">(' + unseen + ')</span>' :
            elt.readAttribute('l'));
    },

    getFolderId: function(f) {
        return 'fld' + decodeURIComponent(f).replace(/_/g,'__').replace(/[^0-9a-z\-_:\.]/gi, '_');
    },

    getSubFolderId: function(f) {
        return 'sub' + f;
    },

    // Preview pane functions
    togglePreviewPane: function() {
        DMenu.close();
        this.showPreview = !this.showPreview;
        $('previewtoggle').setText(this.showPreview ? DimpText.hide_preview : DimpText.show_preview);
        if (this.showPreview) {
            $('msgList').removeClassName('msglistNoPreview');
        } else {
            $('msgList').addClassName('msglistNoPreview');
        }
        new Ajax.Request(DimpVar.URI_PREFS, { parameters: { app: 'dimp', pref: 'show_preview', value: this.showPreview ? 1 : 0 } });
        this.viewport.showSplitPane(this.showPreview);
        if (this.showPreview) {
            this.initPreviewPane();
        }
        this.updateTitle();
    },

    /* Folder list updates. */
    pollFolders: function() {
        // Reset poll folder counter.
        this.setPollFolders();
        var args = $H();
        if (this.folder && $('dimpmain_folder').visible()) {
            args = this.viewport.addRequestParams(args, true);
        }
        $('button_checkmail').setText(DimpText.getmail + ' [' + DimpText.check + ']');
        this.doAction('PollFolders', args, [], this.pollFoldersCallback.bind(this));
    },

    pollFoldersCallback: function(r) {
        if (r.response) {
            r = r.response;
            $H(r.poll).keys().each(function(f) {
                this.setFolderLabel(f, r.poll[f].u);
                if (this.viewport) {
                    this.viewport.setMetaData({ unseen: r.poll[f].u }, f);
                }
            }.bind(this));
        }
        $('button_checkmail').setText(DimpText.getmail);
    },

    setPollFolders: function() {
        if (DimpVar.refresh_time) {
            if (this.pollPE) {
                this.pollPE.stop();
            }
            this.pollPE = new PeriodicalExecuter(this.pollFolders.bind(this), DimpVar.refresh_time);
        }
    },

    showOptions: function(params) {
        this.doAction('ShowOptions', params || { app: 'dimp' }, [], this.portalCallback.bind(this));
        this.portal_effect = Effect.Fade($('dimpmain_portal'));
    },

    portalCallback: function(r) {
        if (this.portal_effect) {
            this.portal_effect.cancel();
            Effect.Appear($('dimpmain_portal'), { duration: .5 });
            this.portal_effect = null;
        }
        if (!r.response) {
            return;
        }
        $('dimpmain_portal').update(r.response.portal, true);
        ToolTips.attachBehavior();
    },

    /* Search filter functions. */
    searchfilterStatus: function() {
        var q = $('qoptions');
        if (q.visible() && !$F('msgList_filter')) {
            this.searchfilterClear();
        } else if ($F('msgList_filter')) {
            q.show();
            this.viewport.onResize();
            this.updateTitle();
        }
    },

    searchfilterOnFocus: function() {
        if (!$('qoptions').visible()) {
            this.setFilterText(false);
        }
    },

    searchfilterOnBlur: function() {
        if (!$('qoptions').visible()) {
            this.setFilterText(true);
        }
    },

    searchfilterClear: function() {
        if (!$('qoptions').visible()) {
            this.setFilterText(true);
            return;
        }
        $('qoptions').hide();
        this.viewport.onResize();
        this.setFilterText(false);
        this.clearSelected();
        this.viewport.runFilter();
    },

    searchfilterClose: function() {
        this.searchfilterClear();
        this.setFilterText(true);
    },

    setFilterText: function(d) {
        var mf = $('msgList_filter');
        if (d) {
            mf.value = DimpText.search;
            mf.addClassName('msgFilterDefault');
        } else {
            mf.value = '';
            mf.removeClassName('msgFilterDefault');
        }
    },

    updateSearchfilter: function(id, type) {
        var c = (type == 'folder') ? this.sfiltersfolder : this.sfilters;
        $H(c).keys().each(function(i) {
            $(i).className = (id == c[i]) ? 'qselected' : '';
        });
        this.viewport.runFilter();
    },

    addSearchfilterParams: function() {
        var sf = $H(this.sfiltersfolder).keys().detect(function(s) {
            return $(s).hasClassName('qselected');
        });
        var sm = $H(this.sfilters).keys().detect(function(s) {
            return $(s).hasClassName('qselected');
        });
        return { searchfolder: this.sfiltersfolder[sf], searchmsg: this.sfilters[sm] };
    },

    /* Add/remove mouse events on the fly.  Current types:
     *   'message', 'draft'  --  Message list rows
     *   'container', 'special', 'folder'  --  Folders
     *   'reply', 'forward', 'otheractions'  --  Message list buttons
     *   'contacts'  --  Linked e-mail addresses */
    addMouseEvents: function(id, type, offset, left, offsettype) {
        try {
            if (!this.drags[id]) {
                switch (type) {
                case 'draft':
                case 'message':
                    this.drags[id] = new DraggableMail(id, DraggableMailOpts);
                    break;

                case 'container':
                case 'folder':
                    this.drags[id] = new Draggable(id, { delay: 100, scroll: 'normalfolders', ghosting: true, revert: true, reverteffect: function(e, t, l) { new Effect.Move(e, { x: -l, y: -t, duration: 0 }); } });
                    break;
                }
            }
            var opts = { offset: offset, left: left, onShow: this.onMenuShow };
            if (offsettype) {
                opts.offsettype = offsettype;
            }
            DMenu.addElement(id, 'ctx_' + type, opts);
        } catch (e) {}
    },

    removeMouseEvents: function(id) {
        if (this.drags[id]) {
            this.drags[id].destroy();
            delete this.drags[id];
        }
        DMenu.removeElement(id);
    },

    onMenuShow: function(ctx) {
        switch (ctx.ctx.substr(4)) {
        case 'folder':
            var folder = $(DMenu.element());
            if (folder.readAttribute('mbox') == 'INBOX') {
                $('ctx_folder_create', 'ctx_folder_rename', 'ctx_folder_delete', 'ctx_folder_sep').invoke('hide');
            } else if (DimpVar.fixed_folders.indexOf(folder.readAttribute('mbox')) != -1) {
                $('ctx_folder_rename', 'ctx_folder_delete').invoke('hide');
            } else {
                $('ctx_folder_create', 'ctx_folder_rename', 'ctx_folder_delete', 'ctx_folder_sep').invoke('show');
            }
            break;
        }
        return true;
    },

    /* Draggable/Droppable handlers. */
    folderHoverHandler: function(elt, folder, overlap) {
        if (overlap > 0.33 && overlap < 0.66 && $(this.getSubFolderId(folder.id))) {
            this.toggleSubFolder(folder.id, 'exp');
        }
    },

    folderDropHandler: function(elt, folder, e) {
        folder = $(folder);
        var elt = $(elt);
        var foldername = folder.readAttribute('mbox');

        if (elt.hasClassName('folder')) {
            // Don't allow drag/drop to INBOX and to child folders (to prevent
            // circular reference issues).
            if (folder.readAttribute('mbox') != 'INBOX' && !folder.descendantOf(this.getSubFolderId(elt.id))) {
                this.doAction('RenameFolder', { old_name: elt.readAttribute('mbox'), new_parent: foldername, new_name: elt.readAttribute('l') }, [], this.folderCallback.bind(this));
            }
            Event.stop(e);
        } else if (folder.readAttribute('ftype') != 'container') {
            var uids;

            if (this.selectedCount() != 0) {
                // Dragging multiple selected messages.
                uids = this.getSelectedRows().pluck('uid');
            } else if (elt.readAttribute('mbox') != foldername) {
                // Dragging a single unselected message.
                uids = [ elt.id.substring(3) ];
            }

            if (uids) {
                this.viewport.updateFlag(uids, 'deleted', true, true);
                this.doAction('MoveMessage', this.viewport.addRequestParams($H({ tofld: foldername }), true), uids, this.deleteCallback.bind(this));
            }
        }
    },

    /* Keypress handler */
    keypressHandler: function(e) {
        // Message list handlers
        if ($('dimpmain_folder').visible()) {
            var kc = e.keyCode || e.charCode;
            switch (kc) {
            case Event.KEY_DELETE:
                var move;
                if (e.shiftKey && this.selectedCount() == 1) {
                    var r = this.viewport.getRowById([ this.selected.first() ]);
                    move = r.rownum + ((r.rownum == this.viewport.getMetaData('total_rows')) ? -1 : 1);
                }
                this.flag('deleted');
                if (move) {
                    this.moveSelected(move, true);
                }
                Event.stop(e);
                break;

            case Event.KEY_UP:
            case Event.KEY_DOWN:
                if (e.shiftKey) {
                    if (this.lastrow) {
                        var uid;
                        switch (kc) {
                        case Event.KEY_UP:
                            if (this.lastrow != 1) {
                                uid = this.viewport.getRows(this.lastrow - 1, 1).first().uid;
                            }
                            break;

                        case Event.KEY_DOWN:
                            if (this.lastrow != this.viewport.getMetaData('total_rows')) {
                                uid = this.viewport.getRows(this.lastrow + 1, 1).first().uid;
                            }
                            break;
                        }

                        if (uid) {
                            this.viewport.positionById(uid);
                            this.shiftSelect(uid);
                        }
                    }
                } else {
                    this.moveSelected(kc == Event.KEY_UP ? -1 : 1);
                }
                Event.stop(e);
                break;

            case Event.KEY_PAGEUP:
            case Event.KEY_PAGEDOWN:
                var ps = this.viewport.getPageSize() - 1;
                var move = ps * (kc == Event.KEY_PAGEUP ? -1 : 1);
                if (this.selectedCount() == 1) {
                    var co = this.viewport.currentOffset();
                    var rowoff = this.viewport.getRowById([ this.selected.first() ]).rownum - 1;
                    switch (kc) {
                    case Event.KEY_PAGEUP:
                        if (co != rowoff) {
                            move = co - rowoff;
                        }
                        break;

                    case Event.KEY_PAGEDOWN:
                        if ((co + ps) != rowoff) {
                            move = co + ps - rowoff;
                        }
                        break;
                    }
                }
                this.moveSelected(move);
                Event.stop(e);
                break;

            case Event.KEY_HOME:
            case Event.KEY_END:
                this.moveSelected(kc == Event.KEY_HOME ? 1 : this.viewport.getMetaData('total_rows'), true);
                Event.stop(e);
                break;

            case Event.KEY_RETURN:
                if (Event.element(e).tagName != 'INPUT') {
                    // Popup message window if single message is selected.
                    var r = this.getSelectedRows();
                    if (r.size() == 1) {
                        r = r.first();
                        this.msgWindow({ uid: r.uid, folder: r.mbox });
                    }
                    Event.stop(e);
                }
                // Don't stop even by default since we may be in an INPUT
                // field, and we want the return to submit the form.
                break;

            case 65: // A
            case 97: // a
                if (e.ctrlKey) {
                    this.selectAll();
                    Event.stop(e);
                }
                break;
            }
        }
    },

    /* Handle rename folder actions. */
    folderrename: function(folder) {
        if (typeof folder == 'undefined') {
            return;
        }
        folder = $(folder);
        DMenu.close();

        var n = this.createFolderForm('renamefolder', function(e) { this.doFolderRename(folder, e); return false; }.bindAsEventListener(this));
        folder.parentNode.insertBefore(n, folder.nextSibling);
        var input = n.down('input');
        input.value = folder.readAttribute('l');
        folder.hide();
        input.focus();
    },

    doFolderRename: function(folder, e) {
        var form = Event.findElement(e, 'form');
        var val = $F(form.down());
        if (val && folder.readAttribute('l') != val) {
            this.doAction('RenameFolder',
                          { old_name: folder.readAttribute('mbox'),
                            new_name: val },
                          [],
                          this.folderCallback.bind(this));
        }
        form.up(1).remove();
        folder.show();
    },

    /* Handle insert folder actions. */
    subfolder: function(folder) {
        if (typeof folder == 'undefined') {
            return false;
        }
        folder = $(folder);
        DMenu.close();

        var n = this.createFolderForm('createfolder', function(e) { this.doSubfolder(folder, e); return false; }.bindAsEventListener(this));
        folder.parentNode.insertBefore(n, folder.nextSibling);
        n.down('input').focus();
    },

    doSubfolder: function(folder, e) {
        var form = Event.findElement(e, 'form');
        var val = $F(form.down());
        if (val) {
            this.doAction('CreateSubfolder',
                          { parent: folder.readAttribute('mbox'),
                            folder: val },
                          [],
                          this.folderCallback.bind(this));
        }
        this.remove_gc = this.remove_gc.concat([ form ]);
        form.up(1).remove();
        folder.show();
    },

    createFolderForm: function(cid, action) {
        var form = $(Builder.node('FORM', { action: '#' }, [ Builder.node('INPUT', { type: 'text', size: 15 }) ]));
        form.observe('submit', action);
        return $(Builder.node('UL', [ Builder.node('LI', { className: cid }, [ form ]) ]));
    },

    /* Folder action callback functions. */
    folderCallback: function(r) {
        if (!r.response) {
            return;
        }

        r.response.d.each(function(i) {
            this.deleteFolder(i);
        }.bind(this));

        r.response.c.each(function(i) {
            this.changeFolder(i);
        }.bind(this));

        r.response.a.each(function(i) {
            this.createFolder(i);
        }.bind(this));
    },

    deleteCallback: function(r) {
        Effect.Fade('folderLoading', { duration: 0.2 });

        if (!r.response ||
            !r.response.uids ||
            !r.response.remove ||
            !r.response.folder == this.folder) {
            return;
        }

        var uids = r.response.uids.parseRangeString();
        this.viewport.remove(uids, r.response.cacheid);
        if (uids.numericSort().toString() == this.selected.numericSort().toString()) {
            this.clearSelected();
            this.toggleButtons();
            this.clearPreviewPane();
        }
    },

    emptyFolderCallback: function(r) {
        if (!r.response.mbox) {
            return;
        }

        if (this.folder == r.response.mbox) {
            this.viewport.reload();
            this.clearPreviewPane();
        }
        this.setFolderLabel(r.response.mbox, 0);
    },

    flagAllCallback: function(r) {
        if (r.response.mbox) {
            this.setFolderLabel(r.response.mbox, r.response.u);
        }
    },

    folderLoadCallback: function(r) {
        if (!r.response) {
            return;
        }

        this.folderCallback(r);

        var fs = $('foldersSidebar');
        fs.observe('click', this.handleFolderMouseEvent.bindAsEventListener(this, 'click'));
        fs.observe('mouseover', this.handleFolderMouseEvent.bindAsEventListener(this, 'over'));
        fs.observe('mouseout', this.handleFolderMouseEvent.bindAsEventListener(this, 'out'));

        $('foldersLoading').hide();
        fs.show();

        var mh = $('normalfolders').getStyle('max-height');

        // Fix for IE6 - which doesn't support max-height.  We need to search
        // for height: 0px instead (comment in IE 6 CSS explains this is
        // needed for auto sizing).
        if (Prototype.Browser.IE && mh === undefined) {
            mh = $('normalfolders').getStyle('height') == '0px' ? true : null;
        }

        if (mh !== null) {
            this.sizeFolderlist();
            Event.observe(window, 'resize', this.sizeFolderlist.bind(this));
        }
    },

    handleFolderMouseEvent: function(e, action) {
        var elt = $(Event.element(e)), li;
        if (elt.nodeType == 3) {
            elt = elt.parentNode;
        }
        if (elt.hasClassName('folder') || elt.hasClassName('custom')) {
            li = elt;
        } else {
            li = elt.up('.folder');
            if (!li) {
                li = elt.up('.custom');
                if (!li) {
                    return;
                }
            }
        }

        var type = li.readAttribute('ftype');

        switch (action) {
        case 'over':
            li.addClassName('over');
            if (type) {
                this.addMouseEvents(li.id, type);
            }
            break;

        case 'out':
            li.removeClassName('over');
            break;

        case 'click':
            if (elt.hasClassName('exp') || elt.hasClassName('col')) {
                this.toggleSubFolder(li.id, 'tog');
            } else {
                switch (type) {
                case 'container':
                    Event.stop(e);
                    break;

                case 'folder':
                case 'special':
                    return this.go('folder:' + li.readAttribute('mbox'));
                    break;
                }
            }
            break;
        }
    },

    toggleSubFolder: function(base, mode) {
        base = $(base);
        var s = $(this.getSubFolderId(base.id));
        if (mode == 'tog' ||
            (mode == 'exp' && !s.visible()) ||
            (mode == 'col' && s.visible())) {
            base.down().className = s.toggle().visible() ? 'col' : 'exp';
        }
    },

    // Folder actions.
    // For format of the ob object, see DIMP::_createFolderElt().
    createFolder: function(ob) {
        var a, div,
            fid = this.getFolderId(ob.m),
            insert_node = 0,
            li, ll,
            mbox = decodeURIComponent(ob.m),
            parent, submbox;

        if (ob.s) {
            parent = $('specialfolders');
        } else {
            parent = $(this.getSubFolderId(this.getFolderId(ob.pa)));
            if (parent) {
                if (parent.down().tagName == 'UL') {
                    parent = parent.down();
                }
            } else {
                parent = $('normalfolders');
            }
        }

        li = $(Builder.node('LI', { className: 'folder', id: fid, l: ob.l, mbox: mbox, ftype: ((ob.co) ? 'container' : 'folder') }));

        div = $(Builder.node('DIV', { className: ob.cl, id: fid + '_div' }));
        if (ob.i) {
            div.update(ob.i);
        }
        if (ob.ch) {
            div.className = 'exp';
        }
        li.appendChild(div);

        a = $(Builder.node('A', { id: fid + '_label', nicetitle: ob.m }, [ ob.l]));
        li.appendChild(a);

        // Now walk through the parent <ul> to find the right place to
        // insert the new folder.
        submbox = $(this.getSubFolderId(fid));
        if (submbox) {
            parent.insertBefore(li, submbox);
            // If an expanded parent mailbox was deleted, we need to toggle the
            // icon accordingly.
            if (submbox.visible()) {
                div.removeClassName('exp');
                div.addClassName('col');
            }
        } else {
            ll = mbox.toLowerCase();
            parent.immediateDescendants().detect(function(node) {
                var nodembox = node.readAttribute('mbox');
                if (nodembox && (!ob.s || nodembox != 'INBOX')) {
                    if (ll < nodembox.toLowerCase()) {
                        insert_node = 1;
                        parent.insertBefore(li, node);
                        return true;
                    }
                }
            });

            if (!insert_node) {
                parent.appendChild(li);
            }
        }

        // Make the new folder a drop target.
        Droppables.add(fid, { hoverclass: 'drop', onHover: this.folderHoverHandler.bind(this), onDrop: this.folderDropHandler.bind(this), scroll: true, overlap: 'vertical' });

        // Need this because add() will make the children 'relative',
        // which breaks the folder list on IE.
        $(fid).undoPositioned();

        // Check for unseen messages
        if (ob.po) {
            li.setAttribute('u', '');
            this.setFolderLabel(mbox, ob.u);
        }

        // Make sure the sub<mbox> ul is created if necessary.
        if (ob.ch && !submbox) {
            parent.insertBefore($(Builder.node('LI', { className: 'subfolders', id: this.getSubFolderId(fid) }, [ Builder.node('UL') ])).hide(), li.nextSibling);
        }
    },

    deleteFolder: function(folder) {
        var f = decodeURIComponent(folder);
        if (this.folder == f) {
            this.go('folder:INBOX');
        }

        var fid = this.getFolderId(folder);
        this.deleteFolderElt(fid, true);

        Effect.Fade(fid, { afterFinish: function(effect) {
            try {
                effect.element.remove();
            } catch (e) {
                this.debug('deleteFolder', e);
            }
        }.bind(this) });
    },

    changeFolder: function(ob) {
        var fid = this.getFolderId(ob.m), fdiv = $(fid + '_div');
        this.deleteFolderElt(fid, !ob.ch);
        var oldexpand = fdiv.hasClassName('col');
        if (ob.co && this.folder == ob.m) {
            this.go('folder:INBOX');
        }
        $(fid).remove();
        this.createFolder(ob);
        if (ob.ch && oldexpand) {
            fdiv.removeClassName('exp');
            fdiv.addClassName('col');
        }
    },

    deleteFolderElt: function(fid, sub) {
        this.remove_gc = this.remove_gc.concat([fid, fid + '_div', fid + '_label']);
        if (sub) {
            var submbox = $(this.getSubFolderId(fid));
            if (submbox) {
                submbox.remove();
            }
        }
        Droppables.remove(fid);
        this.removeMouseEvents(fid);
        if (this.viewport) {
            this.viewport.deleteView(fid);
        }
    },

    sizeFolderlist: function() {
        var fl = $('normalfolders');
        fl.setStyle({ height: (getInnerHeight() - Position.cumulativeOffset(fl)[1] - 10) + 'px' });
    },

    /* Flag actions for message list. */
    flag: function(action, index, folder) {
        var rows = (index && folder)
            ? [{ id: index, uid: index, mbox: folder }]
            : this.getSelectedRows();
        var args, uids = [];

        DMenu.close();

        switch (action) {
        case 'allUnseen':
        case 'allSeen':
            this.doAction((action == 'allUnseen') ? 'MarkFolderUnseen' : 'MarkFolderSeen', { folder: folder }, [], this.flagAllCallback.bind(this));
            if (folder == this.folder) {
                this.viewport.updateFlag(this.viewport.getRows(1, this.viewport.getBufferSize()).pluck('uid'), 'unseen', action == 'allUnseen', true);
            }
            break;

        case 'deleted':
        case 'undeleted':
        case 'spam':
        case 'ham':
        case 'blacklist':
        case 'whitelist':
            if (!rows.size()) {
                break;
            }
            args = this.viewport.addRequestParams($H(), true);
            uids = rows.pluck('uid');
            if (action == 'deleted' || action == 'undeleted') {
                this.viewport.updateFlag(uids, 'deleted', action == 'deleted', true);
            }

            if (action == 'undeleted') {
                this.doAction('UndeleteMessage', args, uids);
            } else {
                var actionCall = { deleted: 'DeleteMessage', spam: 'ReportSpam', ham: 'ReportHam', blacklist: 'Blacklist', whitelist: 'Whitelist' };
                // This needs to be synchronous Ajax if we are calling from a
                // popup window because Mozilla will not correctly call the
                // callback function if the calling window has been closed.
                this.doAction(actionCall[action], args, uids, this.deleteCallback.bind(this), { asynchronous: !(index && folder) });

                // If reporting spam, to indicate to the user that something is
                // happening (since spam reporting may not be instantaneous.
                if (action == 'spam' || action == 'ham') {
                    Effect.Appear('folderLoading', { duration: 0.2 });
                }
            }
            break;

        case 'unseen':
        case 'seen':
            if (!rows.size()) {
                break;
            }
            args = { folder: this.folder, messageFlag: '-seen' };
            var unseenstatus = 1;
            if (action == 'seen') {
                unseenstatus = 0;
                args.messageFlag = 'seen';
            }
            rows.each(function(s) {
                if (this.updateUnseenUID(s.uid, s.mbox, unseenstatus)) {
                    uids.push(s.uid);
                }
            }.bind(this));

            if (uids.length) {
                this.doAction('MarkMessage', args, uids);
            }
            break;

        case 'flagged':
        case 'clear':
            if (!rows.size()) {
                break;
            }
            args = {
                folder: this.folder,
                messageFlag: ((action == 'flagged') ? 'flagged' : '-flagged')
            };
            this.viewport.updateFlag(rows.pluck('uid'), 'flagged', action == 'flagged', true);
            this.doAction('MarkMessage', args, uids);
            break;

        case 'answered':
            uids = rows.pluck('uid');
            this.viewport.updateFlag(uids, 'answered', true, true);
            this.viewport.updateFlag(uids, 'flagged', false, true);
            break;
        }

        return false;
    },

    /* Add a popdown menu to a dimpactions button. */
    addPopdown: function(bid, ctx, offsettype) {
        var bidelt = $(bid),
            img = $($('popdown_img').down().cloneNode(false));
        img.id = bid + '_img';
        var parentelt = bidelt.up();
        parentelt.appendChild(img);
        parentelt.observe('mouseover', this.setPopdownTimeout.bind(this, parentelt, bidelt.next().id));
        parentelt.observe('mouseout', function() { clearTimeout(this.currentButton); }.bind(this));
        bidelt.observe('click', function() { clearTimeout(this.currentButton); }.bind(this));
        this.addMouseEvents(img.id, ctx, parentelt, true, offsettype);
    },

    setPopdownTimeout: function(p, id) {
        this.currentButton = setTimeout(function() {
            if (!p.hasClassName('disabled') && !DMenu.currentmenu()) {
                DMenu.trigger(id, true);
            }
        }, 1000);
    },

    /* Add dropdown menus to address links. */
    buildAddressLinks: function(id) {
        var acount = 0;
        $(id).getElementsByClassName('address').each(function(e) {
            e.id = 'addr' + acount++;
            e.observe('click', Event.stop);
            e.observe('mouseover', function() { this.removeMouseEvents(e.id); this.addMouseEvents(e.id, 'contacts', e, true, 'positioned') }.bind(this));
        }.bind(this));
    }

};

/* Stuff to do after window loads. */
Event.observe(window, 'load', function() {
    /* Set up context menu handler. */
    if (typeof ContextSensitive != 'undefined') {
        DMenu = new ContextSensitive();
    }

    /* Don't do additional onload stuff if we are in a popup. */
    if (parent.opener && parent.opener.DimpCore &&
        parent.opener.location.host == window.location.host) {
        baseWindow = parent.opener.baseWindow || parent.opener;
        return;
    } else if (!$('dimpLoading')) {
        /* We have directly accessed a popup page (e.g. compose.php,
         * message.php). */
        return;
    }

    /* Create the folder list. Any pending notifications will be caught via
     * the return from this call. */
    this.doAction('ListFolders', $H(), [], this.folderLoadCallback.bind(this));

    /* Catch shift-selects in IE - only needed for message list page. */
    if (Prototype.Browser.IE) {
        Event.observe(document, 'selectstart', function(e) { if (DimpVar.is_ie6 || !Event.element(e).descendantOf('previewPane')) { Event.stop(e); } });
    }

    /* Remove unneeded buttons. */
    if (!DimpVar.spam_reporting) {
        this.buttons = this.buttons.without('button_spam');
    }
    if (!DimpVar.ham_reporting) {
        this.buttons = this.buttons.without('button_ham');
    }

    /* Check for new mail. */
    this.setPollFolders();

    $('dimpLoading').hide();
    $('dimpPage').show();

    if (dhtmlHistory.initialize()) {
        dhtmlHistory.addListener(this.show.bind(this));
    }

    /* Initialize the starting page if necessary. addListener() will have
     * already fired if there is a current location so only do a show() call
     * if there is no current location. */
    if (!dhtmlHistory.getCurrentLocation()) {
        this.show('portal');
    }

    /* Load quota bar. */
    if (DimpVar.show_quota) {
        this.checkQuota();
        new PeriodicalExecuter(this.checkQuota.bind(this), DimpVar.refresh_time);
    }

    /* Bind keypresses to document. */
    Event.observe(document, 'keypress', this.keypressHandler.bind(this));

    /* Init garbage collection function - runs every 10 seconds. */
    new PeriodicalExecuter(function() {
        if (this.remove_gc.size()) {
            Event.stopAllObservers(this.remove_gc.splice(0, 75));
            if (!this.remove_gc.size()) {
                Event.observers = Event.observers.compact();
            }
        }
    }.bind(this), 10);

    /* In preview pane, clear any context menus on mousewheel scroll - it is
     * the only case we can not catch elsewhere. */
    $('previewPane').observe(Prototype.Browser.IE ? 'mousewheel' : 'DOMMouseScroll', DMenu.close);

    /* Add popdown menus. */
    this.addPopdown('button_reply', 'reply', 'cumulative');
    DMenu.disable('button_reply_img', true, true);
    this.addPopdown('button_forward', 'forward', 'cumulative');
    DMenu.disable('button_forward_img', true, true);
    this.addPopdown('button_other', 'otheractions', 'cumulative');
    DMenu.setNoClose('button_other');
}.bind(DimpCore));

/* We need to replace the IMP javascript for this function with code that
 * calls the correct DIMP functions. */
function popup_imp(url, w, h, args) {
    var params = $H();
    args.split('&').each(function(pair) {
        pair = pair.split('=');
        params[pair[0]] = pair[1];
    });
    DimpCore.compose('new', params);
}

/* Necessary includes for toggle divs. */
function toggleQuoteBlock(id, lines) {
    dimpToggleBlock($('qb_' + id), $('qt_' + id), DimpText.hidetext, DimpText.showtext + ' - ' + lines + ' ' + DimpText.lines, 'fontSize:70%;');
}

function toggleAddressesBlock(field, count) {
    dimpToggleBlock($('ab_' + field), $('at_' + field), DimpText.hideaddr, DimpText.showaddr + ' - ' + count + ' ' + DimpText.recips);
}

function dimpToggleBlock(block, toggle, hidetxt, showtxt, style) {
    block.toggle();

    var link = $(Builder.node('A', { className: 'widget', style: style || '' }));
    link.observe('click', function() { dimpToggleBlock(block, toggle, hidetxt, showtxt, style); return false; });

    link.appendChild(document.createTextNode('[' + (block.visible() ? hidetxt : showtxt) + ']'));

    if (toggle.firstChild) {
        $(toggle.firstChild).remove();
    }
    toggle.appendChild(link);
}

/* For attachment viewing to work, replaces the functions from
 * horde/templates/contents/open_view_win.js and imp/js/unblockImages.js. */
function view(url, partid) {
    window.open(url, ++DimpCore.compose_id + partid.replace(/\W/g, '_'), 'menubar=yes,toolbar=no,location=no,status=no,scrollbars=yes,resizable=yes');
}

function IMPUnblockImages(parent, message) {
    if (!parent) {
        return true;
    }

    parent.descendants().each(function(elt) {
        if (elt.hasAttribute('blocked')) {
            var src = decodeURIComponent(elt.readAttribute('blocked'));
            if (elt.hasAttribute('src')) {
                elt.setAttribute('src', src);
            } else if (elt.hasAttribute('background')) {
                elt.setAttribute('background', src);
            } else if (elt.hasAttribute('background-image')) {
                elt.setAttribute('background-image', src);
            }
        }
    });

    message = $(message);
    if (message) {
        message.remove();
    }

    // On success return false to stop event propagation.
    return false;
}

/* General utility function. */
function getInnerHeight() {
    var de = document.documentElement;
    return window.innerHeight ? window.innerHeight : (de.clientHeight ? de.clientHeight : document.body.clientHeight);
}

/* Helper methods for setting/getting element text without mucking
 * around with multiple TextNodes. */
Element.addMethods({
    setText: function(element, text) {
        element = $(element);
        if (!element) {
            return;
        }
        var t = 0;
        $A(element.childNodes).each(function(node) {
            if (node.nodeType == 3) {
                if (t === 0) {
                    node.nodeValue = text;
                } else {
                    node.remove();
                }
                ++t;
            }
        });
    },

    getText: function(element, recursive) {
        element = $(element);
        if (!element) {
            return;
        }
        var text = '';
        $A(element.childNodes).each(function(node) {
            if (node.nodeType == 3) {
                text += node.nodeValue;
            } else if (recursive) {
                text += node.getText(true);
            }
        });
        return text;
    }
});

/* Stop all observers for a given element(s). */
Object.extend(Event, {
    stopAllObservers: function(elts) {
        if (!Event.observers) {
            return;
        }
        elts.each(function(elt) {
            elt = $(elt);
            for (var i = 0, length = Event.observers.size(); i < length; ++i) {
                if (Event.observers[i] && Event.observers[i][0] == elt) {
                    Event.stopObserving.apply(Event, Event.observers[i]);
                    delete Event.observers[i];
                }
            }
        });
    }
});

/* Create some utility functions. */
Object.extend(Array.prototype, {
    toRangeString: function() {
        var first = null, last = null, out = [];
        this.numericSort().each(function(k) {
            if (first) {
                if (last + 1 == k) {
                    last = k;
                } else {
                    out.push(first + (last == first ? '' : (':' + last)));
                    first = last = k;
                }
            } else {
                first = last = k;
            }
        });

        out.push(first + (last == first ? '' : (':' + last)));
        return out.join(',');
    },

    numericSort: function() {
        return this.sort(function(a,b) { return a - b; });
    },

    intersect: function(a) {
        return a.findAll(function(e) {
            return this.indexOf(e) != -1;
        }.bind(this));
    }
});

Object.extend(String.prototype, {
    parseRangeString: function() {
        var mlist = [], str = this.strip();
        if (!str.blank()) {
            str.split(',').each(function(e) {
                var r = e.split(':');
                if (r.size() == 1) {
                    mlist.push(parseInt(e));
                } else {
                    mlist = mlist.concat($A($R(parseInt(r[0]), parseInt(r[1]))));
                }
            });
        }
        return mlist;
    },

    // We define our own version of evalScripts() to make sure that all
    // scripts are running in the same scope and that all functions are
    // defined in the global scope. This is not the case when using
    // prototype's evalScripts().
    evalScripts: function() {
        var funcs, scripts = this.extractScripts(), re = /function\s+([^\s(]+)/g;
        for (var i = 0, length = scripts.length; i < length; ++i) {
            eval(scripts[i]);
            while (func = re.exec(scripts[i])) {
                window[func[1]] = eval(func[1]);
            }
        }
    }
});

Object.extend(Hash.prototype, {
    // TODO: Remove once incorporated into prototypejs.
    // See http://dev.rubyonrails.org/ticket/8528 - This appears to have been
    // incorporated into Prototypejs 1.5.2
    keyOf: function(object) {
        var elt = this.find(function(pair) {
            return pair.value == object;
        });
        return elt === undefined ? elt : elt.key;
    },

    flip: function() {
        var h = $H();
        this.each(function(l) {
            h[l.value] = l.key;
        });
        return h;
    }
});
