/**
 * compose.js - Javascript code used in the DIMP compose view.
 *
 * $Horde: dimp/js/src/compose.js,v 1.53 2007/07/12 05:42:25 slusarz Exp $
 *
 * Copyright 2005-2007 The Horde Project (http://www.horde.org/)
 *
 * See the enclosed file COPYING for license information (GPL). If you
 * did not receive this file, see http://www.fsf.org/copyleft/gpl.html.
 */

var auto_save_interval,
    auto_save_interval_val,
    button_pressed,
    editor_on = false,
    identities,
    last_msg = '',
    mp_padding = 0,
    popup_window,
    qreply,
    row_height,
    show_editor = false,
    uploading;

function confirmCancel() {
    if (window.confirm(DimpText_compose.cancel)) {
        if (auto_save_interval_val) {
            DimpCore.doAction('DeleteDraft', { index: $F('index') });
        }
        if (popup_window) {
            if (baseWindow) {
                DimpCore.closePopup();
            } else {
                window.location = DimpVar.URI_DIMP_INBOX;
            }
        } else {
            closeQReply();
            return true;
        }
    }
    return false;
}

function closeQReply() {
    last_msg = '';
    var al = $('attach_list').immediateDescendants();
    if (al.size()) {
        removeAttach(al);
    }
    $('messageCache').value = $('draft_index').value = '';
    $('qreply', 'sendcc', 'sendbcc').invoke('hide');
    $('msglistplacer').show();
    if (editor_on) {
        toggleHtmlEditor();
    }
    Form.reset('compose');

    // Disable auto-save-drafts now.
    if (auto_save_interval) {
        window.clearInterval(auto_save_interval);
        auto_save_interval = null;
    }
}

function change_identity() {
    var id = $F('identity');
    var idnum = (editor_on) ? 1 : 0,
        last = get_identity($F('last_identity')),
        msgval = $('message'),
        next = get_identity(id),
        ssm = $('save_sent_mail');

    $('sent_mail_folder_label').setText(next.id[5]);
    $('bcc').value = next.id[6];
    if (ssm) {
        ssm.checked = next.id[4];
    }

    // Finally try and replace the signature.
    var msg;
    if (editor_on) {
        switch (DimpVar_compose.js_editor) {
        case 'xinha':
            msg = _editors.message.getHTML();
            break;

        case 'fckeditor':
            msg = FCKeditorAPI.GetInstance('message').GetHTML();
            break;

        case 'tinymce':
            msg = tinyMCE.getInstanceById('message').getHTML();
            break;
        }
    } else {
        msg = msgval.value;
    }
    msg = msg.replace(/\r\n/g, '\n');

    var pos = (last.id[2])
        ? msg.indexOf(last.sig)
        : msg.lastIndexOf(last.sig);

    if (pos != -1) {
        if (next.id[2] == last.id[2]) {
            msg = msg.substring(0, pos) + next.sig + msg.substring(pos + last.sig.length, msg.length);
        } else if (next.id[2]) {
            msg = next.sig + msg.substring(0, pos) + msg.substring(pos + last.sig.length, msg.length);
        } else {
            msg = msg.substring(0, pos) + msg.substring(pos + last.sig.length, msg.length) + next.sig;
        }

        msg = msg.replace(/\r\n/g, '\n').replace(/\n/g, '\r\n');
        if (editor_on) {
            switch (DimpVar_compose.js_editor) {
            case 'xinha':
                _editors.message.setHTML(msg);
                break;

            case 'fckeditor':
                FCKeditorAPI.GetInstance('message').SetHTML(msg);
                break;

            case 'tinymce':
                tinyMCE.getInstanceById('message').setHTML(msg);
                break;
            }
        } else {
            msgval.value = msg;
        }
        $('last_identity').value = id;
    }
}

function get_identity(id) {
    return { id: identities[id], sig: identities[id][((editor_on) ? 1 : 0)].replace(/^\n/, '') };
}

function uniqueSubmit(action) {
    var c = $('compose');
    if (action == 'send_message' || action == 'save_draft') {
        c.setStyle({ cursor: 'wait' });
        button_pressed = true;

        switch (action) {
        case 'send_message':
            var sb = $('send_button');

            if (sb.visible()) {
                sb.hide();
                sb.parentNode.insertBefore(Builder.node('LABEL', { id: 'send_label' }, DimpText_compose.sending), sb);
            }
            break;

        case 'save_draft':
            var db = $('draft_button');

            if (db.visible()) {
                db.hide();
                db.parentNode.insertBefore(Builder.node('LABEL', { id: 'draft_label' }, DimpText_compose.saving), db);
            }
            break;
        }

        // Don't send/save until uploading is completed.
        if (uploading) {
            setTimeout(function() { if (button_pressed) { uniqueSubmit(action); } }, 250);
            return;
        }
    }
    $('action').value = action;

    if (action == 'add_attachment') {
        // We need a submit action here because browser security models won't
        // let us access files on user's filesystem otherwise.
        uploading = true;
        c.submit();
    } else {
        // Move HTML text to textarea field for submission.
        if (editor_on) {
            switch (DimpVar_compose.js_editor) {
            case 'xinha':
                // This onsubmit() is needed because xinha sets an onsubmit
                // handler to set the value of the textarea, and since we are
                // overloading the default onsubmit handler, we need to make
                // sure we explicitly call the function.
                c.onsubmit();
                break;

            case 'fckeditor':
                FCKeditorAPI.GetInstance('message').UpdateLinkedField();
                break;

            case 'tinymce':
                tinyMCE.getInstanceById('message').triggerSave();
                break;
            }
        }

        // Use an AJAX submit here so that we can do javascript-y stuff before
        // having to close the window on success.
        DimpCore.doAction('*' + DimpVar.compose_url, c.serialize(true), [], uniqueSubmitCallback);
    }
}

function uniqueSubmitCallback(r) {
    var d = r.response;
    if (!d) {
        return;
    }

    if (d.imp_compose) {
        $('messageCache').value = d.imp_compose;
    }

    if (d.success || d.action == 'add_attachment') {
        switch (d.action) {
        case 'auto_save_draft':
            button_pressed = false;
            $('draft_index').value = d.draft_index;
            break;

        case 'save_draft':
            button_pressed = false;
            if (popup_window && DimpVar_compose.close_draft) {
                baseWindow.DimpCore.showNotifications(r.msgs);
                DimpCore.closePopup();
                return;
            }
            break;

        case 'send_message':
            button_pressed = false;
            /* If we loaded compose.php directly (without a base DIMP window
             * DIMP window), then redirect to the base page. */
            if (!baseWindow) {
                if (qreply) {
                    closeQReply();
                } else {
                    window.location = DimpVar.URI_DIMP_INBOX + '#folder:INBOX';
                }
            } else {
                if (d.reply_type == 'reply') {
                    baseWindow.DimpCore.flag('answered', d.reply_index, d.reply_folder);
                }

                if (d.folder) {
                    baseWindow.DimpCore.createFolder(d.folder);
                }

                if (popup_window) {
                    baseWindow.DimpCore.showNotifications(r.msgs);
                    DimpCore.closePopup();
                    return;
                } else if (qreply) {
                    closeQReply();
                }
            }
            break;

        case 'add_attachment':
            var elt;
            uploading = false;
            if (d.success) {
                addAttach(d.info.number, d.info.name, d.info.type, d.info.size);
            } else {
                button_pressed = false;
            }
            if (DimpVar_compose.attach_limit != -1 &&
                $('attach_list').immediateDescendants().size() > DimpVar_compose.attach_limit) {
                $('upload').disabled = false;
                elt = $(Builder.node('DIV', [ DimpText_compose.attachment_limit ]));
            } else {
                elt = $(Builder.node('INPUT', { type: 'file', name: 'file_1' }));
                elt.observe('change', uploadAttachment);
            }
            elt.id = 'upload';
            $('attach_cell').replaceChild(elt, $('upload_wait'));
            resizeTextarea();
            break;
        }
    } else {
        button_pressed = false;
    }

    // Re-enable buttons if needed.
    if (!button_pressed) {
        var db = $('draft_button'), sb = $('send_button');
        if (!sb.visible()) {
            $('send_label').remove();
            sb.show();
        }
        if (!db.visible()) {
            $('draft_label').remove();
            db.show();
        }
    }

    DimpCore.showNotifications(r.msgs);
}

function toggleHtmlEditor(noupdate) {
    if (!DimpVar_compose.rte_avail) {
        return;
    }
    noupdate = noupdate || false;
    if (SpellCheckerObject) {
        SpellCheckerObject.resume();
    }
    if (editor_on) {
        var text;
        editor_on = false;
        Event.observe(window, 'resize', resizeTextarea);

        switch (DimpVar_compose.js_editor) {
        case 'xinha':
            var msg = _editors.message;
            text = msg.getHTML();
            $(msg._htmlArea).remove();
            msg = _editors = null;
            break;

        case 'fckeditor':
            Event.stopObserving(window, 'resize', resizeFCKeditor);
            var oEditor = FCKeditorAPI.GetInstance('message');
            text = oEditor.GetHTML();
            $('messageParent').immediateDescendants().invoke('hide');
            $('message').show();
            break;

        case 'tinymce':
            Event.stopObserving(window, 'resize', resizeTinyMCEeditor);
            text = tinyMCE.getInstanceById('message').getHTML();
            tinyMCE.execCommand('mceRemoveControl', false, 'message');
            break;
        }

        DimpCore.doAction('Html2Text', { text: text }, [], setMessageText, { asynchronous: false });
    } else {
        editor_on = true;
        // We need to unregister because Xinha defines its own event handlers
        // that sometimes interferes with what we want to do.
        Event.stopObserving(window, 'resize', resizeTextarea);
        if (!noupdate) {
            DimpCore.doAction('Text2Html', { text: $('message').value }, [], setMessageText, { asynchronous: false });
        }

        switch (DimpVar_compose.js_editor) {
        case 'xinha':
            xinha_init();
            break;

        case 'fckeditor':
            oFCKeditor.Height = getTextAreaHeight();
            // Try to reuse the old fckeditor instance.
            try {
                var oEditor = FCKeditorAPI.GetInstance('message');
                oEditor.SetHTML($('message').value);
                $('messageParent').immediateDescendants().invoke('show');
                $('message').hide();
            } catch (e) {
                oFCKeditor.ReplaceTextarea();
            }
            Event.observe(window, 'resize', resizeFCKeditor);
            break;

        case 'tinymce':
            // TinyMCE automatically resizes on init based on current size of
            // messageParent.
            tinyMCE.execCommand('mceAddControl', false, 'message');
            Event.observe(window, 'resize', resizeTinyMCEeditor);
            break;
        }
    }
    $('htmlcheckbox').checked = editor_on;
    $('html').value = editor_on ? 1 : 0;
}

function resizeFCKeditor() {
    $(oFCKeditor.InstanceNameIframe).setStyle({ height: getTextAreaHeight() + 'px' });
}

function resizeTinyMCEeditor() {
    var ed = tinyMCE.getInstanceById('message'),
        h = getTextAreaHeight(),
        t = $(ed.editorId + '_parent').down();
    var diff = h - t.getHeight();
    t.setStyle({ height: h + 'px' });
    ed.iframeElement.setStyle({ height: (ed.iframeElement.getHeight() + diff) + 'px' });
}

function getTextAreaHeight() {
    return getInnerHeight() - Position.cumulativeOffset($('messageParent'))[1] - mp_padding;
}

function initializeSpellChecker() {
    if (!DimpVar_compose.rte_avail) {
        return;
    }

    if (typeof SpellCheckerObject != 'object') {
        // If we fired before the onload that initializes the
        // spellcheck, wait.
        setTimeout(initializeSpellChecker, 100);
        return;
    }

    SpellCheckerObject.onBeforeSpellCheck = function() {
        if (!editor_on) {
            return;
        }
        SpellCheckerObject.htmlAreaParent = 'messageParent';
        switch (DimpVar_compose.js_editor) {
        case 'xinha':
            SpellCheckerObject.htmlArea = _editors.message._htmlArea;
            _editors.message._textArea.value = _editors.message.outwardHtml(_editors.message.getHTML());
            break;

        case 'fckeditor':
            SpellCheckerObject.htmlArea = $(oFCKeditor.InstanceNameIframe);
            $('message').value = FCKeditorAPI.GetInstance('message').GetHTML();
            break;

        case 'tinymce':
            var ob = tinyMCE.getInstanceById('message');
            SpellCheckerObject.htmlArea = $(ob.editorId + '_parent');
            $('message').value = ob.getHTML();
            break;
        }
    };
    SpellCheckerObject.onAfterSpellCheck = function() {
        if (!editor_on) {
            return;
        }
        SpellCheckerObject.htmlArea = SpellCheckerObject.htmlAreaParent = null;
        switch (DimpVar_compose.js_editor) {
        case 'xinha':
            _editors.message.setHTML(_editors.message.inwardHtml(_editors.message._textArea.value));
            break;

        case 'fckeditor':
            FCKeditorAPI.GetInstance('message').SetHTML($('message').value);
            break;

        case 'tinymce':
            tinyMCE.getInstanceById('message').setHTML($('message').value);
            break;
        }
    };
}

function setMessageText(r) {
    var ta = $('message');
    if (!ta) {
        ta = $(Builder.node('TEXTAREA', { id: 'message', name: 'message', style: 'width:100%;' }, r.response.text));
        $('messageParent').appendChild(ta);
    } else {
        ta.value = r.response.text;
    }

    if (!editor_on) {
        resizeTextarea();
    }
}

function fillForm(msg, header, focus, noupdate) {
    var msgval = $('message');

    if (last_msg != '' && last_msg != msgval.value.replace(/\r/g, '') &&
        !window.confirm(DimpText_compose.fillform)) {
        return;
    }

    // Set auto-save-drafts now if not already active.
    if (auto_save_interval_val && !auto_save_interval) {
        auto_save_interval = window.setInterval(function() {
            var cur_msg;
            if (editor_on) {
                switch (DimpVar_compose.js_editor) {
                case 'xinha':
                    cur_msg = _editors.message.getHTML();
                    break;

                case 'fckeditor':
                    cur_msg = FCKeditorAPI.GetInstance('message').GetHTML();
                    break;

                case 'tinymce':
                    cur_msg = tinyMCE.getInstanceById('message').getHTML();
                    break;
                }
            } else {
                cur_msg = msgval.value;
            }
            cur_msg = cur_msg.replace(/\r/g, '');
            if (cur_msg != '' && last_msg != cur_msg) {
                uniqueSubmit('auto_save_draft');
                last_msg = cur_msg;
            }
        }, auto_save_interval_val * 60000);
    }

    if (editor_on) {
        switch (DimpVar_compose.js_editor) {
        case 'xinha':
            _editors.message.setHTML(msg);
            last_msg = _editors.message.getHTML();
            break;

        case 'fckeditor':
            var fo = FCKeditorAPI.GetInstance('message');
            fo.SetHTML(msg);
            last_msg = fo.GetHTML();
            break;

        case 'tinymce':
            var control = tinyMCE.getInstanceById('message');
            control.setHTML(msg);
            last_msg = control.getHTML();
            break;
        }
        last_msg = last_msg.replace(/\r/g, '');
    } else {
        msgval.value = msg;
        last_msg = msgval.value.replace(/\r/g, '');
    }

    $('to').value = header.to;
    if (header.cc) {
        $('cc').value = header.cc;
    }
    if (DimpVar_compose.cc) {
        $('togglecc').checked = true;
        $('sendcc').show();
    }
    if (header.bcc) {
        $('bcc').value = header.bcc;
    }
    if (DimpVar_compose.bcc) {
        $('togglebcc').checked = true;
        $('sendbcc').show();
    }
    $('subject').value = header.subject;
    $('in_reply_to').value = header.in_reply_to;
    $('references').value = header.references;
    $('reply_type').value = header.replytype;

    Field.focus(focus || 'to');
    resizeTextarea();

    if (show_editor) {
        if (!editor_on) {
            toggleHtmlEditor(noupdate || false);
        }
        if (focus == 'message') {
            focusEditor();
        }
    }
}

function focusEditor() {
    try {
        switch (DimpVar_compose.js_editor) {
        case 'xinha':
            _editors.message.focusEditor();
            break;

        case 'fckeditor':
            FCKeditorAPI.GetInstance('message').Focus();
            break;

        case 'tinymce':
            tinyMCE.execCommand('mceFocus', false, 'message');
            break;
        }
    } catch (e) {
        setTimeout('focusEditor()', 100);
    }
}

function addAttach(number, name, type, size) {
    var div = $(Builder.node('DIV', [ name + ' [' + type + '] (' + size + ' KB) ' ]));

    var input = $(Builder.node('INPUT', { type: 'button', atc_id: number, value: DimpText_compose.remove }));
    input.observe('click', function() { removeAttach([ input.up() ]); });
    div.appendChild(input);
    $('attach_list').appendChild(div);
    resizeTextarea();
}

function removeAttach(e) {
    var ids = [];
    e.each(function(n) {
        n = $(n);
        ids.push(n.down().readAttribute('atc_id'));
        n.remove();
    });
    DimpCore.doAction('DeleteAttach', { atc_indices: ids, imp_compose: $F('messageCache') });
    resizeTextarea();
}

function resizeTextarea() {
    var msg = $('message');

    if (!row_height) {
        var m = msg.cloneNode(false);
        // Change the ID and name to not conflict with msg node.
        m.id = m.name = null;
        m.setStyle({ visibility: 'hidden' });
        document.body.appendChild(m);
        m.setAttribute('rows', 1);
        row_height = m.getHeight();
        m.setAttribute('rows', 2);
        row_height = m.getHeight() - row_height;
        m.remove();

        // For some reason, Safari requires us to define this value in here.
        // It's Safari, so don't try to figure out why this is the case.
        mp_padding = $('messageParent').getHeight() - msg.getHeight();
    }

    /* Logic: Determine the size of a given textarea row, divide that size by
     * the available height, round down to the lowest integer row, and resize
     * the textarea. */
    msg.setAttribute('rows', parseInt(getTextAreaHeight() / row_height));
    // IE 7 needs this.
    msg.disabled = false;
}

function uploadAttachment() {
    var u = $('upload');
    $('submit_frame').observe('load', attachmentComplete);
    uniqueSubmit('add_attachment');
    Event.stopObserving(u, 'change', uploadAttachment);
    var div = $(Builder.node('DIV', { id: 'upload_wait' }, [ DimpText_compose.uploading + ' ' + u.value ]));
    $('attach_cell').replaceChild(div, u);
}

function attachmentComplete() {
    var sf = $('submit_frame');
    sf.stopObserving('load', attachmentComplete);
    var doc = sf.contentDocument || sf.contentWindow.document;
    DimpCore.doActionComplete({ responseText: doc.body.innerHTML }, uniqueSubmitCallback);
}

Event.observe(window, 'load', function() {
    initializeSpellChecker();
    $('upload').observe('change', uploadAttachment);

    // Various events that may cause the textarea to grow larger than the
    // window size
    $('togglecc').observe('click', resizeTextarea);
    $('togglebcc').observe('click', resizeTextarea);
    Event.observe(window, 'resize', resizeTextarea);

    // Safari requires a submit target iframe to be at least 1x1 size or else
    // it will open content in a new window.  See:
    //   http://blog.caboo.se/articles/2007/4/2/ajax-file-upload
    if (Prototype.Browser.WebKit) {
        var sf = $('submit_frame');
        sf.setAttribute('position', 'absolute');
        sf.setAttribute('width', '1px');
        sf.setAttribute('height', '1px');
        sf.setStyle({ left: '-999px' });
        sf.show();
    }
});
